// Decacharger Firmware V1.0
// Written By: Patrick McCabe

#include <Wire.h>
#include <LiquidCrystal.h>
#include "ina.h"

#define BUTTON_1 A2    // reset button pin for USB port 1
#define BUTTON_2 A1    // reset button pin for USB port 2
#define BUTTON_3 A0    // reset button pin for USB bank 3
#define BUTTON_4 8     // reset button pin for USB bank 4
#define PRESSED  HIGH  // button inputs read high when pressed

#define VUSB_CTRL_1 12 // on and off control pin for USB port 1
#define VUSB_CTRL_2 11 // on and off control pin for USB port 2
#define VUSB_CTRL_3 10 // on and off control pin for USB bank 3
#define VUSB_CTRL_4 13 // on and off control pin for USB bank 4

#define TURN_ON  LOW      // a low signal turns a port on
#define TURN_OFF HIGH     // a high signal turns a port off

#define TEMP_SENS      A3 // input pin for temperature sensor
#define CURRENT_SENS_3 A6 // input pin for current sense on the 1 amp USB bank
#define CURRENT_SENS_4 A7 // input pin for current sense on the 2.1 amp USB bank

#define FAN_CTRL 9        // PWM control pin for the cooling fan

#define SAFETY_REFRESH_TIME 50          // the time, in ms, between updates of the safety shutoff system
#define LCD_REFRESH_TIME    400         // the time, in ms, between updates of the display
#define BUTTON_REFRESH_TIME 200         // the time, in ms, between reads of the reset buttons
#define TOTAL_CURRENT_REFRESH_TIME 2000 // the time, in ms, between displays of the total current
#define TOTAL_CURRENT_DISPLAY_TIME 2000 // the time, in ms, that the total current is displayed

#define DEBUG false // debug mode - true/false

// the vusb control pins in an array
int vUSBCtrlPins[] = {VUSB_CTRL_1, VUSB_CTRL_2, VUSB_CTRL_3, VUSB_CTRL_4};

const int MON_ADDRESS = 0x40;  // the ina address
ina monitor(MON_ADDRESS, AVG_16, CT_332, CT_332); // the ina current sensor object

LiquidCrystal lcd(2,3,4,5,6,7); // LCD object

unsigned long lastSafetyRefresh = 0;         // stores the last time the safety systems were refreshed
unsigned long lastLCDRefresh = 0;            // stores the last time the lcd was refreshed
unsigned long lastButtonRefresh = 0;         // stores the last time the reset buttons were read
unsigned long lastTotalCurrentRefresh = 0;   // stores the last time the total current was displayed

boolean startedDisplayingTotalCurrent = false;
unsigned long startTimeDisplayingTotalCurrent = 0;

float temp = 0;     // stores the temperature in F
boolean tempWentAboveTurnOnThreshold = false;

float currents[6];  // stores the currents when read



void setup(){
  // set the serial baud rate if debug mode is on
  if(DEBUG){
    Serial.begin(9600); 
  }
  
  initPins();        // set the pinmodes for the input and output pins
  lcd.begin(16,2);   // setup the lcd - 16 columns, 2 rows
  Wire.begin();      // setup I2C
  setupMonitor();    // setup the ina current monitor
  powerUpAllPorts(); // turn all of the usb ports on
  // setting the fan ctrl pin as an output pulls it low and turns the fan on
  // so turn the fan off and leave it up to the temp regulation function
  digitalWrite(FAN_CTRL, TURN_OFF);
}



void initPins(){
  const int NUM_INPUTS = 7;  // the number of inputs
  const int NUM_OUTPUTS = 5; // the number of outputs
  // all the input pins in an array
  int inputs[] = {BUTTON_1, BUTTON_2, BUTTON_3, BUTTON_4, TEMP_SENS, CURRENT_SENS_3, CURRENT_SENS_4};
  // all of the output pins in an array
  int outputs[] = {VUSB_CTRL_1, VUSB_CTRL_2, VUSB_CTRL_3, VUSB_CTRL_4, FAN_CTRL};
  
  for(int i=0; i < NUM_INPUTS; i++){  // iterate through all of the inputs
    pinMode(inputs[i],INPUT);         // set the mode of each input pin as an input
    delay(1);                         // a necessary delay... reason unknown
  }
  for(int i=0; i < NUM_OUTPUTS; i++){ // iterate through all of the outputs
    pinMode(outputs[i], OUTPUT);      // set the mode of each output pin as an output
    delay(1);                         // a necessary delay... reason unknown
  }
}



void powerUpAllPorts(){   // turns the power on for each USB bank
  for(int i=0; i<4; i++){ // iterate through all the ports
    turnPortOn(i);        // turn each one on
  }
}



void setupMonitor(){
  int i;
  for(i = 0; i < 4; i++){
    int err = monitor.setMode(CHAN_1 | CHAN_2 | CHAN_3, SHUNT_MEAS);
    if(err){
      if(DEBUG){
        Serial.print("INA fail: ");
        Serial.println(err);
      }
      // arbitrary, assuming TWI problem
      delay(20);
    }
    else{
      break;
    }
  }

  if(DEBUG){
    if(i<4){
      Serial.print("INA3221 brought up after ");
      Serial.print(i + 1);
      Serial.println(" attempt(s).");
    } 
    else{
      Serial.println("Failed to bring up INA3221.");
    }
  }
}



void loop(){
  unsigned long currentTime = millis();
  if(currentTime - lastSafetyRefresh >= SAFETY_REFRESH_TIME){ // runs the safety systems
    temp = calculateTemp();
    tempRegulation();
    readCurrents();
    safetyShutOff();
    lastSafetyRefresh = millis();
  }
  
  if(currentTime - lastLCDRefresh >= LCD_REFRESH_TIME){ // refreshes the LCD with new data
    displayOutput();
    lastLCDRefresh = millis();
  }
  
  if(currentTime - lastButtonRefresh >= BUTTON_REFRESH_TIME){ // reads the button inputs
    resetUpdate();
    lastButtonRefresh = millis();
  }
 
}




float calculateTemp(){
  const float VIN = 5.0;     // voltage input of the voltage divider
  const float R2 = 10000.0;  // resistance of the resistor in the voltage divider
  const float R0 = 10000.0;  // resistance of ntc at 25C
  const float B = 3977.0;    // beta parameter of ntc
  const float T0 = 298.15;   // 25C in Kelvin

  // implementing the beta parameter equation 
  float Vout = (analogRead(TEMP_SENS) / 1024.0) * 5.0;     // the voltage out of the voltage divider
  float ntcResistance = (VIN*R2 - Vout*R2) / Vout;         // calculating the resistance of the ntc
  float Rinf = R0*exp(-B/T0);                              // calculating R infinity
  float tempKelvin = (B/log(ntcResistance/Rinf));          // the temperature in kelvin 
  float tempFahrenheit = ((tempKelvin-273.15)*1.8) + 32.0; // the temperature in fahrenheit
  
  if(DEBUG){
    Serial.println();
    Serial.print("Temperature: ");
    Serial.print(tempFahrenheit,1);
    Serial.println("F");
  }
  //TODO use real units... pfft
  return tempFahrenheit;
}



void tempRegulation(){
  // the temperature, in F, at which the cooling fan is turned on
  // this temperature is measured on one of the heatsinks
  const int TEMP_TURN_ON = 100;
  // the temperature, in F, at which the cooling fan is turned off
  const int TEMP_TURN_OFF = 90;
  
  // turns the fan on if the temperature is greater than the turn_on threshold
  if(temp > TEMP_TURN_ON && tempWentAboveTurnOnThreshold == false){
    digitalWrite(FAN_CTRL, TURN_ON);
    tempWentAboveTurnOnThreshold = true;
  }
  
  // turns the fan off if the temperature falls below the turn_off threshold
  if(temp < TEMP_TURN_OFF && tempWentAboveTurnOnThreshold == true){
    digitalWrite(FAN_CTRL, TURN_OFF);
    tempWentAboveTurnOnThreshold = false;
  }
}



void resetUpdate(){ // resets USB ports when the reset buttons are pushed
  // all the button digital input pins in an array
  int buttonPins[] = {BUTTON_1, BUTTON_2, BUTTON_3, BUTTON_4};

  for (int i=0; i<4; i++){ // iterate through the button array
    // checks to see if any of the buttons were pressed
    if(digitalRead(buttonPins[i]) == PRESSED){
      turnPortOn(i); // turn the port on
      
      // I don't know why, but this break fixed an interesting issue
      break; // register only one button press per call - seems reasonable 
    }
  }
}



void readCurrents(){
  float reading = 0;
  
  if(monitor.trigger() != 0){
    // there was an error triggering - better luck next time
  }
  else{
    reading = monitor.getVal(CHAN_1, SHUNT_MEAS); // read the shunt voltage on vusb1
    // if the reading is a valid one then convert it to current and store it
    if(validReading(reading)){
      currents[0] = ((reading*3.3) / 0.001)/1000000.0; // USB port 1 current draw
    }
    
    reading = monitor.getVal(CHAN_2, SHUNT_MEAS); // read the shunt voltage on vusb2
    // if the reading is a valid one then convert it to current and store it
    if(validReading(reading)){
      currents[1] = ((reading*1.3) / 0.001)/1000000.0; // USB port 2 current draw
    }
    
    currents[2] = (((analogRead(CURRENT_SENS_3)/1024.0) * 5) / 0.005); // USB bank 3 current draw
    currents[3] = (((analogRead(CURRENT_SENS_4)/1024.0) * 5) / 0.005); // USB bank 4 current draw
    
    reading = monitor.getVal(CHAN_3, SHUNT_MEAS); // read the shunt voltge on vusb
    // if the reading is a valid one then convert it to current and store it
    if(validReading(reading)){
      currents[4] = ((reading*1.7) / 0.001)/1000000.0; // VUSB current draw
    }
    
    currents[5] = currents[0] + currents[1] + currents[4]; // total current draw
  }
  
  if(DEBUG){
    Serial.println();
    Serial.println("Current being drawn:");
    for(int i=0; i<6; i++){
      Serial.print("Current " + String(i) + ": ");
      Serial.println(currents[i],2);
    }
  }
}


// checks to see if an ina reading is valid or not
boolean validReading(float valFloat){
  // a valid reading is a float that should be less than 1
  // an invalid reading is an integer 1, 2, or 3
  // casting the float to an integer means valid reading will be 0
  int valInt = int(valFloat);
  
  return(valInt == 0 || abs(valInt) >= 4);
}


void safetyShutOff(){ 
  const float MAX_CURRENT_1 = 5.0;      // maximum current draw allowed on USB port 1
  const float MAX_CURRENT_2 = 5.0;      // maximum current draw allowed on USB port 2
  const float MAX_CURRENT_3 = 7.0;      // maximum current draw allowed on USB bank 3
  const float MAX_CURRENT_4 = 6.0;      // maximum current draw allowed on USB bank 4
  const float MAX_CURRENT_TOTAL = 15.0; // maximum current draw allowed over everything
  
  // all the maximum currents for each port in an array
  float maxCurrents[] = {MAX_CURRENT_1, MAX_CURRENT_2, MAX_CURRENT_3, MAX_CURRENT_4};
  // keeps track of whether or not a port was turned off
  boolean turnedOffAPort = false; 
  
  // turns off USB ports if they are individually drawing too much current
  for(int i=0; i<4; i++){              // iterate through the individual currents
    if(currents[i] > maxCurrents[i]){  // check to see if the current is greater than its allowed maximum
      turnPortOff(i);                  // if so, turn the port off
      turnedOffAPort = true;           // flag that a port was turned off
    }
  }  
  
  // re-read the currents only if a port was turned off
  if(turnedOffAPort){
    readCurrents();  
  }
  
  // turns off USB ports if the whole system is drawing too much current
  while(currents[5] > MAX_CURRENT_TOTAL){ // loop while the total current being drawn is still too damn high
    int minimumDifference = 100;          // will store the smallest difference between a USB bank and its maximum
    int indexOfPinToShutOff = 0;          // will store the index of the USB port ctrl pin to shut off
    
    for(int i=0; i<4; i++){  // iterate through the individual currents
      // the difference between the maximum current allowed and the current on a USB bank (this should be positive)
      int difference = maxCurrents[i] - currents[i]; 
      
      // if the difference is smaller than the current smallest difference
      // and the port is actually drawing some current
      if((difference < minimumDifference) && currents[i] > 0.5){ 
        minimumDifference = difference;   // then set the smallest difference to the current difference
        indexOfPinToShutOff = i;          // store the index of this port
      }
    }
    
    turnPortOff(indexOfPinToShutOff);
    readCurrents();  // re-read the currents after turning a port off
  }
  
}


// takes the index of a ctrl pin and turns that port on
void turnPortOn(int indexOfPort){
  digitalWrite(vUSBCtrlPins[indexOfPort], TURN_ON);
}


// takes the index of a ctrl pin and turns that port off
void turnPortOff(int indexOfPort){
  digitalWrite(vUSBCtrlPins[indexOfPort], TURN_OFF);
}



void displayOutput(){
  // "A: 2.13  B: 3.24" Example of first line of output
  // "   Temp: 87F    " Example of second line of output when displaying temp
  // "   Total: 10.2  " Example of second line of output when displaying total current
  
  lcd.clear();           
  lcd.home();                
  
  lcd.print("A: ");       
  lcd.print(currents[0],2);  
  lcd.print("  ");
  lcd.print("B: ");
  lcd.print(currents[1],2);
  

  // displays the total current on the second line of the LCD
  // if enough time has elapsed since the last time it was displayed
  // and if it hasn't been on the screen for too long since first displaying
  if(millis() - lastTotalCurrentRefresh >= TOTAL_CURRENT_REFRESH_TIME &&
    (millis() - startTimeDisplayingTotalCurrent <= TOTAL_CURRENT_DISPLAY_TIME ||
     startTimeDisplayingTotalCurrent == 0)){
       
    if(startedDisplayingTotalCurrent == false){  // if this is the first time through
      startTimeDisplayingTotalCurrent = millis();// record the start time of displaying the total current
      startedDisplayingTotalCurrent = true;  
    }
    
    lcd.setCursor(0,1);
    lcd.print("   Total: ");
    lcd.print(currents[5],2);
  }
  // displays the temperature otherwise
  else{
    lcd.setCursor(0,1);
    lcd.print("   Temp: ");
    lcd.print(int(temp));
    lcd.print("F");
    
  }
  
  // resets the last time the total current was displayed
  // once it finishes displaying for the defined amount of time
  if((millis() - startTimeDisplayingTotalCurrent > TOTAL_CURRENT_DISPLAY_TIME) && startTimeDisplayingTotalCurrent != 0){
    lastTotalCurrentRefresh = millis();
    startedDisplayingTotalCurrent = false;
    startTimeDisplayingTotalCurrent = 0;
  }
  
}




